@@ -41,16 +41,28 @@
 #include "multiplot.h"
 
 #include "command.h"
+#include "datablock.h"
 #include "gadgets.h"
 #include "graphics.h"
+#include "misc.h"
 #include "parse.h"
 #include "setshow.h"
 #include "util.h"
 
-
+/* Local prototypes */
 static void mp_layout_size_and_offset(void);
 static void mp_layout_margins_and_spacing(void);
 static void mp_layout_set_margin_or_spacing(t_position *);
+static void init_multiplot_datablock(void);
+
+/* Support for multiplot playback */
+TBOOLEAN multiplot_playback = FALSE;	  /* TRUE while inside "remultiplot" playback */
+TBOOLEAN suppress_multiplot_save = FALSE; /* TRUE inside a for/while loop */
+static t_value multiplot_udv = {
+	.type = DATABLOCK,
+	.v.data_array = NULL
+};
+int multiplot_last_panel = 0;
 
 enum set_multiplot_id {
     S_MULTIPLOT_LAYOUT,
@@ -199,8 +211,7 @@ multiplot_start()
     c_token++;
 
     /* Only a few options are possible if we are already in multiplot mode */
-    /* So far we have "next".  Maybe also "previous", "clear"? */
-    if (multiplot) {
+    if (in_multiplot > 0) {
 	if (equals(c_token, "next")) {
 	    c_token++;
 	    if (!mp_layout.auto_layout)
@@ -408,14 +419,15 @@ multiplot_start()
     }
 
     /* If we reach here, then the command has been successfully parsed.
-     * Aug 2013: call term_start_plot() before setting multiplot so that
+     * Call term_start_plot() before setting multiplot so that
      * the wxt and qt terminals will reset the plot count to 0 before
      * ignoring subsequent TERM_LAYER_RESET requests. 
      */
     term_start_plot();
-    multiplot = TRUE;
+    in_multiplot = lf_head ? lf_head->depth + 1 : 1;
     multiplot_count = 0;
     fill_gpval_integer("GPVAL_MULTIPLOT", 1);
+    init_multiplot_datablock();
 
     /* Place overall title before doing anything else */
     if (mp_layout.title.text) {
@@ -453,7 +465,7 @@ multiplot_start()
 void
 multiplot_end()
 {
-    multiplot = FALSE;
+    in_multiplot = 0;
     multiplot_count = 0;
     fill_gpval_integer("GPVAL_MULTIPLOT", 0);
     /* reset plot size, origin and margins to values before 'set
@@ -484,9 +496,30 @@ multiplot_end()
 	free(mp_layout.title.text);
 	mp_layout.title.text = NULL;
     }
+
+    if (!multiplot_playback) {
+	/* Create or recycle a user-visible datablock for replay */
+	struct udvt_entry *datablock = add_udv_by_name("$GPVAL_LAST_MULTIPLOT");
+	free_value(&datablock->udv_value);
+
+	/* Add the closing "unset multiplot" command before saving */
+	append_to_datablock(&multiplot_udv, strdup("unset multiplot"));
+
+	datablock->udv_value = multiplot_udv;
+	multiplot_udv.v.data_array = NULL;
+
+	/* Save panel number of last-drawn plot */
+	multiplot_last_panel = mp_layout.current_panel;
+    }
+    last_plot_was_multiplot = TRUE;
+}
+
+/* Helper functions for multiplot auto layout to issue size and offset cmds */
+TBOOLEAN multiplot_auto()
+{
+    return mp_layout.auto_layout_margins;
 }
 
-/* Helper function for multiplot auto layout to issue size and offset cmds */
 void
 multiplot_reset()
 {
@@ -619,3 +652,56 @@ mp_layout_set_margin_or_spacing(t_position *margin)
     }
 }
 
+/* Begin accummulating lines that can later be replayed by replay_multiplot().
+ * The first line to be stored is a comment.
+ * Subsequent lines will be appended as they are read in.
+ */
+static void
+init_multiplot_datablock()
+{
+    gpfree_datablock(&multiplot_udv);
+    multiplot_udv.type = DATABLOCK;
+    append_to_datablock(&multiplot_udv, strdup("# saved multiplot"));
+}
+
+/* Append one line to the multiplot history.
+ * Called from two places:
+ *	command.c:com_line() catches direct input from stdin
+ *	misc.c:load_file() catches lines from load/call
+ * When the multiplot is exited via multiplot_end(), all lines will be
+ * copied to the user-visible datblock $GPVAL_LAST_MULTIPLOT.
+ */
+void
+append_multiplot_line(char *line)
+{
+    if (line == NULL || *line == '\0')
+	return;
+    if (*line == '$' && strstr(line,"<<"))
+	return;
+    if (suppress_multiplot_save)
+	return;
+
+    append_to_datablock(&multiplot_udv, strdup(line));
+}
+
+/* This is the implementation of "remultiplot".
+ * The load operation is protected by a flag to prevent infinite recursion.
+ */
+void
+replay_multiplot()
+{
+    multiplot_playback = TRUE;
+    load_file(NULL, strdup("$GPVAL_LAST_MULTIPLOT"), 6);
+    multiplot_playback = FALSE;
+}
+
+/* reset needed if int_error() is invoked during multiplot playback */
+void
+multiplot_reset_after_error()
+{
+    if (!multiplot_playback)
+	return;
+    multiplot_end();
+    multiplot_playback = FALSE;
+    suppress_multiplot_save = FALSE;
+}
